@@ -15,9 +15,12 @@
 package requestbody
 
 import (
+	"time"
+
+	"github.com/dustin/go-humanize"
+
 	"github.com/caddyserver/caddy/v2/caddyconfig/httpcaddyfile"
 	"github.com/caddyserver/caddy/v2/modules/caddyhttp"
-	"github.com/dustin/go-humanize"
 )
 
 func init() {
@@ -25,25 +28,48 @@ func init() {
 }
 
 func parseCaddyfile(h httpcaddyfile.Helper) (caddyhttp.MiddlewareHandler, error) {
+	h.Next() // consume directive name
+
 	rb := new(RequestBody)
 
-	for h.Next() {
-		// configuration should be in a block
-		for h.NextBlock(0) {
-			switch h.Val() {
-			case "max_size":
-				var sizeStr string
-				if !h.AllArgs(&sizeStr) {
-					return nil, h.ArgErr()
-				}
-				size, err := humanize.ParseBytes(sizeStr)
-				if err != nil {
-					return nil, h.Errf("parsing max_size: %v", err)
-				}
-				rb.MaxSize = int64(size)
-			default:
-				return nil, h.Errf("unrecognized servers option '%s'", h.Val())
+	// configuration should be in a block
+	for h.NextBlock(0) {
+		switch h.Val() {
+		case "max_size":
+			var sizeStr string
+			if !h.AllArgs(&sizeStr) {
+				return nil, h.ArgErr()
 			}
+			size, err := humanize.ParseBytes(sizeStr)
+			if err != nil {
+				return nil, h.Errf("parsing max_size: %v", err)
+			}
+			rb.MaxSize = int64(size)
+
+		case "read_timeout":
+			var timeoutStr string
+			if !h.AllArgs(&timeoutStr) {
+				return nil, h.ArgErr()
+			}
+			timeout, err := time.ParseDuration(timeoutStr)
+			if err != nil {
+				return nil, h.Errf("parsing read_timeout: %v", err)
+			}
+			rb.ReadTimeout = timeout
+
+		case "write_timeout":
+			var timeoutStr string
+			if !h.AllArgs(&timeoutStr) {
+				return nil, h.ArgErr()
+			}
+			timeout, err := time.ParseDuration(timeoutStr)
+			if err != nil {
+				return nil, h.Errf("parsing write_timeout: %v", err)
+			}
+			rb.WriteTimeout = timeout
+
+		default:
+			return nil, h.Errf("unrecognized request_body subdirective '%s'", h.Val())
 		}
 	}
 
