@@ -19,15 +19,19 @@ import (
 	"crypto/x509"
 	"encoding/base64"
 	"encoding/json"
+	"encoding/pem"
 	"fmt"
 	"io"
 	"os"
 	"path/filepath"
 	"strings"
 
-	"github.com/caddyserver/caddy/v2"
-	"github.com/mholt/acmez"
+	"github.com/mholt/acmez/v2"
 	"go.uber.org/zap"
+
+	"github.com/caddyserver/caddy/v2"
+	"github.com/caddyserver/caddy/v2/caddyconfig"
+	"github.com/caddyserver/caddy/v2/caddyconfig/caddyfile"
 )
 
 func init() {
@@ -54,7 +58,7 @@ func (cp ConnectionPolicies) Provision(ctx caddy.Context) error {
 		}
 
 		// enable HTTP/2 by default
-		if len(pol.ALPN) == 0 {
+		if pol.ALPN == nil {
 			pol.ALPN = append(pol.ALPN, defaultALPN...)
 		}
 
@@ -115,6 +119,9 @@ func (cp ConnectionPolicies) TLSConfig(_ caddy.Context) *tls.Config {
 						continue policyLoop
 					}
 				}
+				if pol.Drop {
+					return nil, fmt.Errorf("dropping connection")
+				}
 				return pol.TLSConfig, nil
 			}
 
@@ -152,6 +159,9 @@ type ConnectionPolicy struct {
 	// Maximum TLS protocol version to allow. Default: `tls1.3`
 	ProtocolMax string `json:"protocol_max,omitempty"`
 
+	// Reject TLS connections. EXPERIMENTAL: May change.
+	Drop bool `json:"drop,omitempty"`
+
 	// Enables and configures TLS client authentication.
 	ClientAuthentication *ClientAuthentication `json:"client_authentication,omitempty"`
 
@@ -159,6 +169,18 @@ type ConnectionPolicy struct {
 	// is no policy configured for the empty SNI value.
 	DefaultSNI string `json:"default_sni,omitempty"`
 
+	// FallbackSNI becomes the ServerName in a ClientHello if
+	// the original ServerName doesn't match any certificates
+	// in the cache. The use cases for this are very niche;
+	// typically if a client is a CDN and passes through the
+	// ServerName of the downstream handshake but can accept
+	// a certificate with the origin's hostname instead, then
+	// you would set this to your origin's hostname. Note that
+	// Caddy must be managing a certificate for this name.
+	//
+	// This feature is EXPERIMENTAL and subject to change or removal.
+	FallbackSNI string `json:"fallback_sni,omitempty"`
+
 	// Also known as "SSLKEYLOGFILE", TLS secrets will be written to
 	// this file in NSS key log format which can then be parsed by
 	// Wireshark and other tools. This is INSECURE as it allows other
@@ -216,6 +238,7 @@ func (p *ConnectionPolicy) buildStandardTLSConfig(ctx caddy.Context) error {
 				cfg.CertSelection = p.CertSelection
 			}
 			cfg.DefaultServerName = p.DefaultSNI
+			cfg.FallbackServerName = p.FallbackSNI
 			return cfg.GetCertificate(hello)
 		},
 		MinVersion: tls.VersionTLS12,
@@ -270,7 +293,7 @@ func (p *ConnectionPolicy) buildStandardTLSConfig(ctx caddy.Context) error {
 			break
 		}
 	}
-	if !alpnFound {
+	if !alpnFound && (cfg.NextProtos == nil || len(cfg.NextProtos) > 0) {
 		cfg.NextProtos = append(cfg.NextProtos, acmez.ACMETLS1Protocol)
 	}
 
@@ -287,8 +310,10 @@ func (p *ConnectionPolicy) buildStandardTLSConfig(ctx caddy.Context) error {
 
 	// client authentication
 	if p.ClientAuthentication != nil {
-		err := p.ClientAuthentication.ConfigureTLSConfig(cfg)
-		if err != nil {
+		if err := p.ClientAuthentication.provision(ctx); err != nil {
+			return fmt.Errorf("provisioning client CA: %v", err)
+		}
+		if err := p.ClientAuthentication.ConfigureTLSConfig(cfg); err != nil {
 			return fmt.Errorf("configuring TLS client authentication: %v", err)
 		}
 	}
@@ -303,7 +328,7 @@ func (p *ConnectionPolicy) buildStandardTLSConfig(ctx caddy.Context) error {
 			return err
 		}
 		logFile, _, err := secretsLogPool.LoadOrNew(filename, func() (caddy.Destructor, error) {
-			w, err := os.OpenFile(filename, os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0600)
+			w, err := os.OpenFile(filename, os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0o600)
 			return destructableWriter{w}, err
 		})
 		if err != nil {
@@ -340,12 +365,18 @@ func (p ConnectionPolicy) SettingsEmpty() bool {
 
 // ClientAuthentication configures TLS client auth.
 type ClientAuthentication struct {
+	// Certificate authority module which provides the certificate pool of trusted certificates
+	CARaw json.RawMessage `json:"ca,omitempty" caddy:"namespace=tls.ca_pool.source inline_key=provider"`
+	ca    CA
+
+	// DEPRECATED: Use the `ca` field with the `tls.ca_pool.source.inline` module instead.
 	// A list of base64 DER-encoded CA certificates
 	// against which to validate client certificates.
 	// Client certs which are not signed by any of
 	// these CAs will be rejected.
 	TrustedCACerts []string `json:"trusted_ca_certs,omitempty"`
 
+	// DEPRECATED: Use the `ca` field with the `tls.ca_pool.source.file` module instead.
 	// TrustedCACertPEMFiles is a list of PEM file names
 	// from which to load certificates of trusted CAs.
 	// Client certificates which are not signed by any of
@@ -354,7 +385,7 @@ type ClientAuthentication struct {
 
 	// DEPRECATED: This field is deprecated and will be removed in
 	// a future version. Please use the `validators` field instead
-	// with the tls.client_auth.leaf module instead.
+	// with the tls.client_auth.verifier.leaf module instead.
 	//
 	// A list of base64 DER-encoded client leaf certs
 	// to accept. If this list is not empty, client certs
@@ -364,7 +395,7 @@ type ClientAuthentication struct {
 	// Client certificate verification modules. These can perform
 	// custom client authentication checks, such as ensuring the
 	// certificate is not revoked.
-	VerifiersRaw []json.RawMessage `json:"verifiers,omitempty" caddy:"namespace=tls.client_auth inline_key=verifier"`
+	VerifiersRaw []json.RawMessage `json:"verifiers,omitempty" caddy:"namespace=tls.client_auth.verifier inline_key=verifier"`
 
 	verifiers []ClientCertificateVerifier
 
@@ -385,13 +416,197 @@ type ClientAuthentication struct {
 	existingVerifyPeerCert func([][]byte, [][]*x509.Certificate) error
 }
 
+// UnmarshalCaddyfile parses the Caddyfile segment to set up the client authentication. Syntax:
+//
+//	client_auth {
+//		mode                   [request|require|verify_if_given|require_and_verify]
+//	 	trust_pool			   <module> {
+//			...
+//		}
+//		trusted_leaf_cert      <base64_der>
+//		trusted_leaf_cert_file <filename>
+//		verifier               <module>
+//	}
+//
+// If `mode` is not provided, it defaults to `require_and_verify` if any of the following are provided:
+// - `trusted_leaf_certs`
+// - `trusted_leaf_cert_file`
+// - `trust_pool`
+//
+// Otherwise, it defaults to `require`.
+func (ca *ClientAuthentication) UnmarshalCaddyfile(d *caddyfile.Dispenser) error {
+	for d.NextArg() {
+		// consume any tokens on the same line, if any.
+	}
+	for nesting := d.Nesting(); d.NextBlock(nesting); {
+		subdir := d.Val()
+		switch subdir {
+		case "mode":
+			if d.CountRemainingArgs() > 1 {
+				return d.ArgErr()
+			}
+			if !d.Args(&ca.Mode) {
+				return d.ArgErr()
+			}
+		case "trusted_ca_cert":
+			caddy.Log().Warn("The 'trusted_ca_cert' field is deprecated. Use the 'trust_pool' field instead.")
+			if len(ca.CARaw) != 0 {
+				return d.Err("cannot specify both 'trust_pool' and 'trusted_ca_cert' or 'trusted_ca_cert_file'")
+			}
+			if !d.NextArg() {
+				return d.ArgErr()
+			}
+			ca.TrustedCACerts = append(ca.TrustedCACerts, d.Val())
+		case "trusted_leaf_cert":
+			if !d.NextArg() {
+				return d.ArgErr()
+			}
+			ca.TrustedLeafCerts = append(ca.TrustedLeafCerts, d.Val())
+		case "trusted_ca_cert_file":
+			caddy.Log().Warn("The 'trusted_ca_cert_file' field is deprecated. Use the 'trust_pool' field instead.")
+			if len(ca.CARaw) != 0 {
+				return d.Err("cannot specify both 'trust_pool' and 'trusted_ca_cert' or 'trusted_ca_cert_file'")
+			}
+			if !d.NextArg() {
+				return d.ArgErr()
+			}
+			filename := d.Val()
+			ders, err := convertPEMFilesToDER(filename)
+			if err != nil {
+				return d.WrapErr(err)
+			}
+			ca.TrustedCACerts = append(ca.TrustedCACerts, ders...)
+		case "trusted_leaf_cert_file":
+			if !d.NextArg() {
+				return d.ArgErr()
+			}
+			filename := d.Val()
+			ders, err := convertPEMFilesToDER(filename)
+			if err != nil {
+				return d.WrapErr(err)
+			}
+			ca.TrustedLeafCerts = append(ca.TrustedLeafCerts, ders...)
+		case "trust_pool":
+			if len(ca.TrustedCACerts) != 0 {
+				return d.Err("cannot specify both 'trust_pool' and 'trusted_ca_cert' or 'trusted_ca_cert_file'")
+			}
+			if !d.NextArg() {
+				return d.ArgErr()
+			}
+			modName := d.Val()
+			mod, err := caddyfile.UnmarshalModule(d, "tls.ca_pool.source."+modName)
+			if err != nil {
+				return d.WrapErr(err)
+			}
+			caMod, ok := mod.(CA)
+			if !ok {
+				return fmt.Errorf("trust_pool module '%s' is not a certificate pool provider", caMod)
+			}
+			ca.CARaw = caddyconfig.JSONModuleObject(caMod, "provider", modName, nil)
+		case "verifier":
+			if !d.NextArg() {
+				return d.ArgErr()
+			}
+
+			vType := d.Val()
+			modID := "tls.client_auth.verifier." + vType
+			unm, err := caddyfile.UnmarshalModule(d, modID)
+			if err != nil {
+				return err
+			}
+
+			_, ok := unm.(ClientCertificateVerifier)
+			if !ok {
+				return d.Errf("module '%s' is not a caddytls.ClientCertificateVerifier", modID)
+			}
+			ca.VerifiersRaw = append(ca.VerifiersRaw, caddyconfig.JSONModuleObject(unm, "verifier", vType, nil))
+		default:
+			return d.Errf("unknown subdirective for client_auth: %s", subdir)
+		}
+	}
+
+	// only trust_ca_cert or trust_ca_cert_file was specified
+	if len(ca.TrustedCACerts) > 0 {
+		fileMod := &InlineCAPool{}
+		fileMod.TrustedCACerts = append(fileMod.TrustedCACerts, ca.TrustedCACerts...)
+		ca.CARaw = caddyconfig.JSONModuleObject(fileMod, "provider", "inline", nil)
+		ca.TrustedCACertPEMFiles, ca.TrustedCACerts = nil, nil
+	}
+	return nil
+}
+
+func convertPEMFilesToDER(filename string) ([]string, error) {
+	certDataPEM, err := os.ReadFile(filename)
+	if err != nil {
+		return nil, err
+	}
+	var ders []string
+	// while block is not nil, we have more certificates in the file
+	for block, rest := pem.Decode(certDataPEM); block != nil; block, rest = pem.Decode(rest) {
+		if block.Type != "CERTIFICATE" {
+			return nil, fmt.Errorf("no CERTIFICATE pem block found in %s", filename)
+		}
+		ders = append(
+			ders,
+			base64.StdEncoding.EncodeToString(block.Bytes),
+		)
+	}
+	// if we decoded nothing, return an error
+	if len(ders) == 0 {
+		return nil, fmt.Errorf("no CERTIFICATE pem block found in %s", filename)
+	}
+	return ders, nil
+}
+
+func (clientauth *ClientAuthentication) provision(ctx caddy.Context) error {
+	if len(clientauth.CARaw) > 0 && (len(clientauth.TrustedCACerts) > 0 || len(clientauth.TrustedCACertPEMFiles) > 0) {
+		return fmt.Errorf("conflicting config for client authentication trust CA")
+	}
+
+	// convert all named file paths to inline
+	if len(clientauth.TrustedCACertPEMFiles) > 0 {
+		for _, fpath := range clientauth.TrustedCACertPEMFiles {
+			ders, err := convertPEMFilesToDER(fpath)
+			if err != nil {
+				return nil
+			}
+			clientauth.TrustedCACerts = append(clientauth.TrustedCACerts, ders...)
+		}
+	}
+
+	// if we have TrustedCACerts explicitly set, create an 'inline' CA and return
+	if len(clientauth.TrustedCACerts) > 0 {
+		clientauth.ca = InlineCAPool{
+			TrustedCACerts: clientauth.TrustedCACerts,
+		}
+		return nil
+	}
+
+	// if we don't have any CARaw set, there's not much work to do
+	if clientauth.CARaw == nil {
+		return nil
+	}
+	caRaw, err := ctx.LoadModule(clientauth, "CARaw")
+	if err != nil {
+		return err
+	}
+	ca, ok := caRaw.(CA)
+	if !ok {
+		return fmt.Errorf("'ca' module '%s' is not a certificate pool provider", ca)
+	}
+	clientauth.ca = ca
+
+	return nil
+}
+
 // Active returns true if clientauth has an actionable configuration.
 func (clientauth ClientAuthentication) Active() bool {
 	return len(clientauth.TrustedCACerts) > 0 ||
 		len(clientauth.TrustedCACertPEMFiles) > 0 ||
 		len(clientauth.TrustedLeafCerts) > 0 || // TODO: DEPRECATED
 		len(clientauth.VerifiersRaw) > 0 ||
-		len(clientauth.Mode) > 0
+		len(clientauth.Mode) > 0 ||
+		clientauth.CARaw != nil || clientauth.ca != nil
 }
 
 // ConfigureTLSConfig sets up cfg to enforce clientauth's configuration.
@@ -420,7 +635,8 @@ func (clientauth *ClientAuthentication) ConfigureTLSConfig(cfg *tls.Config) erro
 		// otherwise, set a safe default mode
 		if len(clientauth.TrustedCACerts) > 0 ||
 			len(clientauth.TrustedCACertPEMFiles) > 0 ||
-			len(clientauth.TrustedLeafCerts) > 0 {
+			len(clientauth.TrustedLeafCerts) > 0 ||
+			clientauth.CARaw != nil {
 			cfg.ClientAuth = tls.RequireAndVerifyClientCert
 		} else {
 			cfg.ClientAuth = tls.RequireAnyClientCert
@@ -428,23 +644,8 @@ func (clientauth *ClientAuthentication) ConfigureTLSConfig(cfg *tls.Config) erro
 	}
 
 	// enforce CA verification by adding CA certs to the ClientCAs pool
-	if len(clientauth.TrustedCACerts) > 0 || len(clientauth.TrustedCACertPEMFiles) > 0 {
-		caPool := x509.NewCertPool()
-		for _, clientCAString := range clientauth.TrustedCACerts {
-			clientCA, err := decodeBase64DERCert(clientCAString)
-			if err != nil {
-				return fmt.Errorf("parsing certificate: %v", err)
-			}
-			caPool.AddCert(clientCA)
-		}
-		for _, pemFile := range clientauth.TrustedCACertPEMFiles {
-			pemContents, err := os.ReadFile(pemFile)
-			if err != nil {
-				return fmt.Errorf("reading %s: %v", pemFile, err)
-			}
-			caPool.AppendCertsFromPEM(pemContents)
-		}
-		cfg.ClientCAs = caPool
+	if clientauth.ca != nil {
+		cfg.ClientCAs = clientauth.ca.CertPool()
 	}
 
 	// TODO: DEPRECATED: Only here for backwards compatibility.
@@ -459,7 +660,7 @@ func (clientauth *ClientAuthentication) ConfigureTLSConfig(cfg *tls.Config) erro
 			}
 			trustedLeafCerts = append(trustedLeafCerts, clientCert)
 		}
-		clientauth.verifiers = append(clientauth.verifiers, LeafCertClientAuth{TrustedLeafCerts: trustedLeafCerts})
+		clientauth.verifiers = append(clientauth.verifiers, LeafCertClientAuth{trustedLeafCerts: trustedLeafCerts})
 	}
 
 	// if a custom verification function already exists, wrap it
@@ -523,17 +724,42 @@ func setDefaultTLSParams(cfg *tls.Config) {
 
 // LeafCertClientAuth verifies the client's leaf certificate.
 type LeafCertClientAuth struct {
-	TrustedLeafCerts []*x509.Certificate
+	LeafCertificateLoadersRaw []json.RawMessage `json:"leaf_certs_loaders,omitempty" caddy:"namespace=tls.leaf_cert_loader inline_key=loader"`
+	trustedLeafCerts          []*x509.Certificate
 }
 
 // CaddyModule returns the Caddy module information.
 func (LeafCertClientAuth) CaddyModule() caddy.ModuleInfo {
 	return caddy.ModuleInfo{
-		ID:  "tls.client_auth.leaf",
+		ID:  "tls.client_auth.verifier.leaf",
 		New: func() caddy.Module { return new(LeafCertClientAuth) },
 	}
 }
 
+func (l *LeafCertClientAuth) Provision(ctx caddy.Context) error {
+	if l.LeafCertificateLoadersRaw == nil {
+		return nil
+	}
+	val, err := ctx.LoadModule(l, "LeafCertificateLoadersRaw")
+	if err != nil {
+		return fmt.Errorf("could not parse leaf certificates loaders: %s", err.Error())
+	}
+	trustedLeafCertloaders := []LeafCertificateLoader{}
+	for _, loader := range val.([]any) {
+		trustedLeafCertloaders = append(trustedLeafCertloaders, loader.(LeafCertificateLoader))
+	}
+	trustedLeafCertificates := []*x509.Certificate{}
+	for _, loader := range trustedLeafCertloaders {
+		certs, err := loader.LoadLeafCertificates()
+		if err != nil {
+			return fmt.Errorf("could not load leaf certificates: %s", err.Error())
+		}
+		trustedLeafCertificates = append(trustedLeafCertificates, certs...)
+	}
+	l.trustedLeafCerts = trustedLeafCertificates
+	return nil
+}
+
 func (l LeafCertClientAuth) VerifyClientCertificate(rawCerts [][]byte, _ [][]*x509.Certificate) error {
 	if len(rawCerts) == 0 {
 		return fmt.Errorf("no client certificate provided")
@@ -544,7 +770,7 @@ func (l LeafCertClientAuth) VerifyClientCertificate(rawCerts [][]byte, _ [][]*x5
 		return fmt.Errorf("can't parse the given certificate: %s", err.Error())
 	}
 
-	for _, trustedLeafCert := range l.TrustedLeafCerts {
+	for _, trustedLeafCert := range l.trustedLeafCerts {
 		if remoteLeafCert.Equal(trustedLeafCert) {
 			return nil
 		}
@@ -573,6 +799,12 @@ type ConnectionMatcher interface {
 	Match(*tls.ClientHelloInfo) bool
 }
 
+// LeafCertificateLoader is a type that loads the trusted leaf certificates
+// for the tls.leaf_cert_loader modules
+type LeafCertificateLoader interface {
+	LoadLeafCertificates() ([]*x509.Certificate, error)
+}
+
 // ClientCertificateVerifier is a type which verifies client certificates.
 // It is called during verifyPeerCertificate in the TLS handshake.
 type ClientCertificateVerifier interface {
@@ -586,3 +818,5 @@ type destructableWriter struct{ *os.File }
 func (d destructableWriter) Destruct() error { return d.Close() }
 
 var secretsLogPool = caddy.NewUsagePool()
+
+var _ caddyfile.Unmarshaler = (*ClientAuthentication)(nil)
